using OrderedCollections
import Base
# Overload Base.parse to deal with arrays of floats
function parseNamelistArray(::Type{T}, s::AbstractString) where T <: Array{DT,1} where DT <: Real
  return parse.(DT,strip.(split(s)))
end

function parseNamelistBool(s::AbstractString)
  if s == "F" || s == "f" || s == ".F." || s == ".f." || s == ".FALSE." || s == ".false."
    return false
  elseif s == "T" || s == "t" || s == ".T." || s == ".t." || s == ".TRUE." || s == ".true."
    return true
  else
    throw(ArgumentError("Could not parse fortran logical"))
  end
end

"""
    readVmecNamelist(nmlFile::AbstractString)

Reads a VMEC input file in Fortran namelist format and populates a corresponding `Dict{Symbol,Any}`,
where the `Symbol` key maps directly to the variable label in the namelist file.  Values the type of
the input variable is defined in the global VmecInputTypes `Dict`.
"""
function readVmecNamelist(nmlFile::AbstractString)
  fileStream = open(nmlFile)
  nmlString = read(fileStream,String)
  close(fileStream)

  # Replace &INDATA and /
  nmlString = replace(nmlString,"&INDATA\n"=>"")
  nmlString = replace(nmlString,"/\n"=>"")

  # Filter the comments
  commentRegex = r"![^\n]*"
  nmlString = replace(nmlString,commentRegex=>"")

  # Filter any entries where the data begins on the next line
  nmlString = replace(nmlString,r"=\s*\n\s*"=>"=")


  # Find locations where the a newline character interrupts data
  newlineArrayLocs = findall(r"[0-9]\s*\n\s*[0-9]",nmlString)
  while !isempty(newlineArrayLocs)
    if !isempty(newlineArrayLocs)
      for r in newlineArrayLocs
        nmlString = nmlString[1:r[1]]*" "*nmlString[r[end]:end]
      end
    end
    newlineArrayLocs = findall(r"[0-9]\s*\n\s*[0-9]",nmlString)
  end

  # Split on new line characters
  nmlLineSplits = split(nmlString,'\n')
  nmlLines = Vector{SubString{String}}(undef,0)

  # Split on any commas
  commaRegex = r"([_*a-zA-Z0-9_*]*\(*\-*[0-9]*,*\-*[0-9]*\)*\s*=\s*[^,]*)"
  for (j,line) in enumerate(nmlLineSplits)
    if !isempty(line)
      matches = collect(eachmatch(commaRegex,line))
      if !isempty(matches)
        push!(nmlLines,strip(matches[1].captures[1]))
        if length(matches) > 1
          for i in 2:length(matches)
            push!(nmlLines,strip(matches[i].captures[1]))
          end
        end
      else
        push!(nmlLines,strip(line))
      end
    end
  end


  nmlDict = OrderedDict{Symbol,Any}()
  for (i,line) in enumerate(nmlLines)
    line2 = replace(line,"="=>" = ")
    matches = collect(eachmatch(r"([^\s]+)",line2))
    @debug "matches = $matches, length(matches) = $(length(matches))"
    @assert length(matches) >= 3 "Error parsing $(line2) in readVmecNamelist2, not a balanced assignment statement"
    key = Symbol(lowercase(matches[1].captures[1]))
    valueStr = ""
    if length(matches) > 3
      valueStr *= "["*sanitizeFortranInput(matches[3].captures[1])
      for j = 4:length(matches)
        valueStr *= ","*sanitizeFortranInput(matches[j].captures[1])
      end
      valueStr *= "]"
    else
      valueStr *= sanitizeFortranInput(matches[3].captures[1])
    end
    nmlDict[key] = eval(Meta.parse(valueStr))
  end
  return nmlDict
end

"""
    sanitizeFortanInput(s::AbstractString)

Return the correct julia type to be used in Meta.parse
"""
function sanitizeFortranInput(s::AbstractString)
  if lowercase(s) == "f" || lowercase(s) == ".false." || lowercase(s) == ".f."
    return "false"
  elseif lowercase(s) == "t" || lowercase(s) == ".true." || lowercase(s) == ".t."
    return "true"
  elseif occursin("'",s)
    return replace(s,"\'"=>"\"")
  else
    return s
  end
end

"""
    updateInputNamelist(nmlDict::AbstractDict{Symbol,Any},vmec::VmecData)

Construct a new dictionary from `nmlDict` with entries updated by the
associated values from `vmec`.

See also: [`updateInputeNamelist!`](@ref)
"""
function updateInputNamelist(namelistDict::AbstractDict{Symbol,Any},
                             vmec::VmecData;
                            )
  newDict = copy(namelistDict)
  updateInputNamelist!(newDict,vmec)
  return newDict
end

"""
    updateInputNamelist!(nmlDict::AbstractDict{Symbol,Any},vmec::VmecData)

Perform an in-place update of the entires of `nmlDict` from the associated
values in `vmec`.

See also: [`updateInputNamelist`](@ref)
"""
function updateInputNamelist!(namelistDict::AbstractDict{Symbol,Any},
                              vmec::VmecData;
                             )
  @assert namelistDict[:mpol] == vmec.mpol "Number of poloidal modes must match"
  @assert namelistDict[:ntor] == vmec.ntor "Number of toroidal modes must match"
  @assert namelistDict[:nfp] == vmec.nfp "Number of field periods must match"
  # Deal with any quantity that has a one-to-one
  # correspondence with the input dictionary
  for key in propertynames(vmec)
    if haskey(VmecInputTypes,key) && isdefined(vmec,key)
      value = getfield(vmec,key)
      inputType = VmecInputTypes[key]
      typeof(getfield(vmec,key)) <: inputType || throw(TypeError("Incorrect type for VMEC namelist parameter $key"))
      namelistDict[key] = value
    end
  end
  # Update the axis and delete the legacy :raxis and :zaxis if necessary
  if haskey(namelistDict,:raxis)
    if !haskey(namelistDict,:raxis_cc)
      typeof(getfield(vmec,:raxis_cc)) <: VmecInputType[:raxis_cc] || throw(TypeError("Incorrect type for VMEC namelist parameter raxis_cc"))
      namelistDict[:raxis_cc] = getfield(vmec,:raxis_cc)
    end
    if !haskey(namelistDict,:raxis_cs)
      typeof(getfield(vmec,:raxis_cs)) <: VmecInputType[:raxis_cs] || throw(TypeError("Incorrect type for VMEC namelist parameter raxis_cs"))
      namelistDict[:raxis_cs] = getfield(vmec,:raxis_cs)
    end
    delete!(namelistDict,:raxis)
  end
  if haskey(namelistDict,:zaxis)
    if !haskey(namelistDict,:zaxis_cc)
      typeof(getfield(vmec,:zaxis_cc)) <: VmecInputType[:zaxis_cc] || throw(TypeError("Incorrect type for VMEC namelist parameter zaxis_cc"))
      namelistDict[:zaxis_cc] = getfield(vmec,:zaxis_cc)
    end
    if !haskey(namelistDict,:zaxis_cs)
      typeof(getfield(vmec,:zaxis_cs)) <: VmecInputType[:zaxis_cs] || throw(TypeError("Incorrect type for VMEC namelist parameter zaxis_cs"))
      namelistDict[:zaxis_cs] = getfield(vmec,:zaxis_cs)
    end
    delete!(namelistDict,:zaxis)
  end
  ntor = vmec.ntor
  mpol = vmec.mpol
  nfp = vmec.nfp
  ns = vmec.ns
  i = 0
  for n = 0:ntor
    i += 1
    sym = Symbol("rbc("*string(n)*",0)")
    namelistDict[sym] = vmec.rmnc[i,ns]
    sym = Symbol("zbs("*string(n)*",0)")
    namelistDict[sym] = vmec.zmns[i,ns]
    if namelistDict[:lasym]
      sym = Symbol("rbs("*string(n)*",0)")
      namelistDict[sym] = Float64[i,ns]
      sym = Symbol("zbc("*string(n)*",0)")
      namelistDict[sym] = Float64[i,ns]
    end
  end
  for m = 1:mpol-1
    for n = -ntor:ntor
      i += 1
      sym = Symbol("rbc("*string(n)*","*string(m)*")")
      namelistDict[sym] = vmec.rmnc[i,ns]
      sym = Symbol("zbs("*string(n)*","*string(m)*")")
      namelistDict[sym] = vmec.zmns[i,ns]
      if namelistDict[:lasym]
        sym = Symbol("rbs("*string(n)*","*string(m)*")")
        namelistDict[sym] = Float64[i,ns]
        sym = Symbol("zbc("*string(n)*","*string(m)*")")
        namelistDict[sym] = Float64[i,ns]
      end
    end
  end
end

"""
    VmecInputTypes <: Dict{Symbol,Type}

Dictionary containing the type information of VMEC input variables.
"""
VmecInputTypes = Dict{Symbol,Type}()

VmecInputTypes[:lasym] = Bool
VmecInputTypes[:lfreeb] = Bool
VmecInputTypes[:lwouttxt] = Bool
VmecInputTypes[:lpofr] = Bool
VmecInputTypes[:full3d1out] = Bool

VmecInputTypes[:ns_array] = Array{Int32,1}
VmecInputTypes[:niter_array] = Array{Int32,1}
VmecInputTypes[:niter] = Int32
VmecInputTypes[:nfp] = Int32
VmecInputTypes[:nstep] = Int32
VmecInputTypes[:mpol] = Int32
VmecInputTypes[:ntor] = Int32
VmecInputTypes[:ncurr] = Int32
VmecInputTypes[:nzeta] = Int32
VmecInputTypes[:nvacskip] = Int32
VmecInputTypes[:nsin] = Int32
VmecInputTypes[:ntheta] = Int32
VmecInputTypes[:mfilter_fbdy] = Int32
VmecInputTypes[:nfilter_fbdy] = Int32
VmecInputTypes[:max_main_iterations] = Int32
VmecInputTypes[:omp_num_threads] = Int32

VmecInputTypes[:ftol_array] = Array{Float64,1}
VmecInputTypes[:ai] = Array{Float64,1}
VmecInputTypes[:am] = Array{Float64,1}
VmecInputTypes[:ac] = Array{Float64,1}
VmecInputTypes[:prec2d_threshold] = Float64
VmecInputTypes[:delt] = Float64
VmecInputTypes[:tcon0] = Float64
VmecInputTypes[:phiedge] = Float64
VmecInputTypes[:extcur] = Float64
VmecInputTypes[:gamma] = Float64
VmecInputTypes[:bloat] = Float64
VmecInputTypes[:spres_ped] = Float64
VmecInputTypes[:pres_scale] = Float64
VmecInputTypes[:curtor] = Float64
VmecInputTypes[:raxis] = Array{Float64,1}
VmecInputTypes[:zaxis] = Array{Float64,1}
VmecInputTypes[:raxis_cc] = Array{Float64,1}
VmecInputTypes[:raxis_cs] = Array{Float64,1}
VmecInputTypes[:zaxis_cc] = Array{Float64,1}
VmecInputTypes[:zaxis_cs] = Array{Float64,1}
VmecInputTypes[:am_aux_s] = Array{Float64,1}
VmecInputTypes[:am_aux_f] = Array{Float64,1}
VmecInputTypes[:ai_aux_s] = Array{Float64,1}
VmecInputTypes[:ac_aux_s] = Array{Float64,1}
VmecInputTypes[:ac_aux_f] = Array{Float64,1}
VmecInputTypes[:aphi] = Array{Float64,1}

VmecInputTypes[:precon_type] = String
VmecInputTypes[:mgrid_file] = String
VmecInputTypes[:pmass_type] = String
VmecInputTypes[:pcurr_type] = String
VmecInputTypes[:piota_type] = String

#Populate the coefficient symbols
for m = 0:32
  for n = -32:32
    sym = Symbol("rbc("*string(n)*","*string(m)*")")
    VmecInputTypes[sym] = Float64
    sym = Symbol("rbs("*string(n)*","*string(m)*")")
    VmecInputTypes[sym] = Float64
    sym = Symbol("zbc("*string(n)*","*string(m)*")")
    VmecInputTypes[sym] = Float64
    sym = Symbol("zbs("*string(n)*","*string(m)*")")
    VmecInputTypes[sym] = Float64
  end
end
