module RotationalTransform
using VMEC
using MPI

function computeTarget(args::Dict{Symbol,Any},vmec::Vmec{T}) where {T <: AbstractFloat}
  return vmec.iota(args[:surfaceLabel])
end

function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T}) where {T <: AbstractFloat}
  if args[:surfaceLabel] != vmecSurface.s
    throw(error("Input surface label and VmecSurface do not match"))
  end
  return vmecSurface.iota[1]
end

function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T},comm::MPI.Comm) where {T <: AbstractFloat}
  return vmecSurface.iota[1]
end

end

module AspectRatio
using VMEC
using MPI

function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T},comm::MPI.Comm) where {T <: AbstractFloat}
  return vmecSurface.aspect
end

end

module MagneticWell
using VMEC
using MPI

function computeTarget(args::Dict{Symbol,Any},
                       vmecSurface::VmecSurface{T},
                       comm::MPI.Comm) where {T <: AbstractFloat}
  return -vmecSurface.vp[2]*vmecSurface.signgs
end

end

module QuasiSymmetry
  using VMEC
  using MPI

  """
      target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,vmec::VmecSurface{T}) where {T <: AbstractFloat}

  Compute the square root of the sum of squares of the non-quasisymmetric modes
  normalized by the (0,0) mode.
  """
  function target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,vmec::VmecSurface{T}) where {T <: AbstractFloat}
    return VMEC.quasisymmetry_deviation(mQS,nQS,mBoozer,nBoozer,vmec)
  end

  function target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,surfaceLabels::AbstractVector{VmecSurface{T}}) where {T}
    targetValues = Vector{Float64}(undef,length(surfaceLabels))
    Threads.@threads for i = 1:length(surfaceLabels)
      targetValues[i] = target_quasisymmetry(mQS,nQS,mBoozer,nBoozer,surfaceLabels[i])
    end
    return sum(targetValues)
  end

  function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T}) where {T}
    if args[:surfaceLabel] != vmecSurface.s
      throw(error("Input surface label and VmecSurface label do not match!"))
    end
    return target_quasisymmetry(args[:mQS],args[:nQS],args[:mBoozer],args[:nBoozer],vmecSurface)
  end

  function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T},comm::MPI.Comm) where {T}
    if MPI.Comm_rank(comm) == 0
      return target_quasisymmetry(args[:mQS],args[:nQS],args[:mBoozer],args[:nBoozer],vmecSurface)
    else
      return nothing
    end
  end

end
